# Copyright (C) 2018-'21  Frank Sachsenheim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

""" This is not an XPath implementation. """

from abc import ABC
from collections import UserString
from functools import lru_cache
from typing import TYPE_CHECKING, Dict, Iterator, List, Optional, Set


if TYPE_CHECKING:
    from delb import TagNode


# ordered by alphabet
AXIS_NAMES = (
    "ancestor",
    "ancestor-or-self",
    "attribute",
    "child",
    "descendant",
    "descendant-or-self",
    "following",
    "following-sibling",
    "namespace",
    "parent",
    "preceding",
    "preceding-sibling",
    "self",
)
# ordered by weight (ascending)
BOOLEAN_OPERATORS = ("or", "and", "!=", "=", ">", ">=", "<", "<=")
BOOLEAN_FUNCTIONS = ("contains", "not", "starts-with")


def _in_parenthesis(expression: str) -> bool:
    """
    Determines whether an expression is surrounded by parenthesis in its
    entirety.

    >>> _in_parenthesis('foo and bar')
    False

    >>> _in_parenthesis('(foo)(bar)')
    False

    >>> _in_parenthesis('foo)') == _in_parenthesis('(foo') == False
    True

    >>> _in_parenthesis('((foo)(bar))')
    True

    """
    if not expression.startswith("(") or not expression.endswith(")"):
        return False
    level = 1
    for character in expression[1:-1]:
        if character == "(":
            level += 1
        elif character == ")":
            level -= 1
        if level < 1:
            return False
    return True


def _partition_terms(expression: str) -> List[str]:  # noqa: C901
    """
    Split up expression into partitions under consideration of quotes and
    parentheses.

    >>> _partition_terms('(@foo="1") and (@bar="2")')
    ['@foo="1"', 'and', '@bar="2"']

    >>> _partition_terms('@foo = "1"')
    ['@foo', '=', '"1"']

    >>> _partition_terms('((@foo="1") or (@bar="2")) and @baz!="3"')
    ['(@foo="1") or (@bar="2")', 'and', '@baz!="3"']

    >>> _partition_terms('@href and starts-with(@href,"https://")')
    ['@href', 'and', 'starts-with(@href,"https://")']

    >>> _partition_terms('@href and not(starts-with(@href,"https://"))')
    ['@href', 'and', 'not(starts-with(@href,"https://"))']

    >>> _partition_terms('not(@foo) and (@bar)')
    ['not(@foo)', 'and', '@bar']

    """
    bracket_level = 0
    function_args = False
    current_term = ""
    quote = ""
    result = []

    for character in expression:
        if quote or bracket_level:
            if character == quote:
                quote = ""
            elif character == "(":
                bracket_level += 1
            elif character == ")":
                bracket_level -= 1
                if not bracket_level and not function_args:
                    continue
                function_args = function_args and bool(bracket_level)
            current_term += character
        # TODO escaped characters
        elif character in ("'", '"'):
            quote = character
            current_term += character
        elif character == "(":
            bracket_level += 1
            if current_term in BOOLEAN_FUNCTIONS:
                current_term += character
                function_args = True
        elif character == ")":
            raise AssertionError
        elif character == " ":
            result.append(current_term)
            current_term = ""
        else:
            current_term += character

    if current_term:
        result.append(current_term)

    return result


def _reduce_whitespace(expression: str) -> str:
    """
    Remove unnecessary whitespace from xpath predicate expression.

    >>> _reduce_whitespace('[@a = "1" or  @b = "2"][@c = "3"]')
    '[@a="1" or @b="2"][@c="3"]'

    >>> _reduce_whitespace('[contains(@a, "1")]')
    '[contains(@a,"1")]'

    """
    quote = ""
    result = ""
    skip = 0

    for i, character in enumerate(expression):
        if skip:
            skip -= 1
        elif character == " " and not quote:
            if expression[i : i + 4] == " or ":
                result += " or "
                skip = 3
            elif expression[i : i + 5] == " and ":
                result += " and "
                skip = 4
            else:
                pass  # result += ""
        elif character in ("'", '"'):
            quote = "" if quote else character
            result += character
        else:
            result += character

    return result


def _split(expression: str, separator: str) -> Iterator[str]:
    """
    Split expression at occurrences of specified seperator, except
    where within quotation marks.

    >>> list(_split('./root/path[@a="n/a"]', '/'))
    ['.', 'root', 'path[@a="n/a"]']

    >>> list(_split('@type="translation" and @xml:lang="en"', ' and '))
    ['@type="translation"', '@xml:lang="en"']

    """
    assert separator not in ('"', "'")
    cursor = 0
    part = ""
    quote = ""
    separator_length = len(separator)

    for i, character in enumerate(expression):
        if i < cursor:
            continue

        if expression[i : i + separator_length] == separator and not quote:
            yield part
            part = ""
            cursor += separator_length
            continue

        if character == quote:
            quote = ""
        elif character in ('"', "'"):
            quote = character

        part += character
        cursor += 1

    yield part


class XPathExpression(UserString):
    __slots__ = ("location_paths",)

    def __init__(self, expression: str):
        self.location_paths = [
            LocationPath(x) for x in _split(_reduce_whitespace(expression), "|")
        ]

    @property
    def data(self):
        return " | ".join(str(x) for x in self.location_paths)

    def is_unambiguously_locatable(self) -> bool:
        """
        Determine whether an xpath expression addresses exactly one possible node.

        Examples::

        >>> XPathExpression('./path1|./path2').is_unambiguously_locatable()
        False

        >>> XPathExpression('/path').is_unambiguously_locatable()
        False

        >>> XPathExpression('./root/node/descendant::foo').is_unambiguously_locatable()
        False

        >>> XPathExpression('./node[@a="1" or @b="2"]').is_unambiguously_locatable()
        False

        >>> XPathExpression('./root/node[@a="1"][@b="2"]').is_unambiguously_locatable()
        True

        """
        if len(self.location_paths) != 1:
            return False

        steps = self.location_paths[0].location_steps

        if steps[0].axis != "self":
            return False

        for step in steps[1:]:
            if step.axis != "child":
                return False

            predicates = step.predicates
            if predicates is None:
                continue

            if not predicates._can_describe_attributes:
                return False

        return True


class LocationPath:
    __slots__ = ("location_steps",)

    def __init__(self, expression: str):
        self.location_steps = [LocationStep(x) for x in _split(expression, "/")]

    def __str__(self):
        return "/".join(str(x) for x in self.location_steps)


class LocationStep:
    __slots__ = ("axis", "node_test", "predicates")

    def __init__(self, expression: str):
        self.axis: str
        self.node_test: NodeTest
        self.predicates: Optional[PredicateExpression] = None

        if expression == "":
            self.axis = "descendant-or-self"
            self.node_test = NodeTest("node()")
        elif expression == ".":
            self.axis = "self"
            self.node_test = NodeTest("node()")
        elif expression == "..":
            self.axis = "parent"
            self.node_test = NodeTest("node()")
        else:
            if "[" in expression:
                # FIXME mind quotes:
                expression, predicates_part = expression.split("[", maxsplit=1)
                assert predicates_part[-1] == "]", predicates_part
                self.predicates = PredicateExpression.parse("[" + predicates_part)

            if expression.startswith("@"):
                expression = "attribute::" + expression[1:]

            if "::" not in expression:
                self.axis = "child"
                self.node_test = NodeTest(expression)
            else:
                self.axis, node_test_part = expression.split("::")
                self.node_test = NodeTest(node_test_part)

        if self.axis not in AXIS_NAMES:
            raise ValueError(f"`{self.axis}` is not a valid axis name.")

    def __str__(self):
        result = self.axis + "::" + str(self.node_test)
        if self.predicates is not None:
            result += "[" + str(self.predicates) + "]"
        return result

    def _derive_attributes(self):
        return self.predicates._derive_attributes()


class NodeTest:
    __slots__ = ("data", "type")

    def __init__(self, expression: str):
        self.data = expression

        if expression.endswith(")"):
            self.type = "type_test"
        else:
            self.type = "name_test"

    def __str__(self):
        return self.data


class PredicateExpression(ABC):
    def evaluate(self, node_set: Set["TagNode"]) -> Set["TagNode"]:
        # shall not be implemented
        raise NotImplementedError

    @staticmethod
    @lru_cache(64)
    def parse(expression: str) -> "PredicateExpression":
        """
        Parse string expression into ``PredicateExpression`` subclass instance.

        >>> isinstance(PredicateExpression.parse('[(@foo)]'), AttributePredicate)
        True

        >>> isinstance(PredicateExpression.parse('[@a="1" or @b="2"]'), BooleanPredicate) # noqa: E501
        True

        >>> isinstance(PredicateExpression.parse('[1]'), IndexPredicate)
        True

        >>> str(PredicateExpression.parse('[@href and starts-with(@href,"https://")]'))
        '(@href and starts-with(@href,"https://"))'

        """
        if expression.startswith("["):
            expression = expression[1:-1]
        expressions = tuple(_split(expression, "]["))

        if len(expressions) > 1:
            parsed_expressions = [
                PredicateExpression.parse(e) for e in expressions[::-1]
            ]

            while len(parsed_expressions) > 1:
                parsed_expressions.append(
                    BooleanPredicate(
                        "and", parsed_expressions.pop(), parsed_expressions.pop()
                    ),
                )
            return parsed_expressions[0]

        while _in_parenthesis(expression):
            expression = expression[1:-1]

        partitions = _partition_terms(expression)

        if len(partitions) == 1:
            if expression.startswith("@") or expression.startswith("attribute::"):
                return AttributePredicate.parse(expression)
            elif expression.isdigit():
                return IndexPredicate(expression)
            elif any(expression.startswith(f) for f in BOOLEAN_FUNCTIONS):
                return FunctionExpression(expression)
            elif expression.startswith('"') or expression.endswith("'"):
                return Literal(expression)
            else:
                return UnsupportedPredicate(expression)

        else:
            return BooleanPredicate.from_partitions(partitions)

    @property
    def _can_describe_attributes(self) -> bool:
        return False

    def _derive_attributes(self) -> Dict[str, str]:
        raise RuntimeError


class AttributePredicate(PredicateExpression):
    __slots__ = ("name", "value")

    def __init__(self, name: str, value: Optional[str]):
        self.name = name
        self.value = value

    @staticmethod
    # related: https://github.com/python/mypy/issues/5107
    def parse(expression: str) -> "AttributePredicate":  # type: ignore
        """
        Parse an xpath predicate string expression into an ``AttributePredicate``
        instance.

        >>> isinstance(
        ...     AttributePredicate.parse('@type="translation"'), AttributePredicate
        ... )
        True

        """
        if expression.startswith("attribute::"):
            raise NotImplementedError
        assert expression.startswith("@")

        parts = tuple(_split(expression[1:], "="))
        if len(parts) == 1:
            return AttributePredicate(parts[0], None)
        elif len(parts) == 2:
            return AttributePredicate(parts[0], parts[1][1:-1])
        else:
            raise ValueError

    def __str__(self):
        value = self.value
        if value is None:
            return f"@{self.name}"
        quote = "'" if '"' in value else '"'
        return f"@{self.name}={quote}{self.value}{quote}"

    @property
    def _can_describe_attributes(self) -> bool:
        return self.value is not None

    def _derive_attributes(self) -> Dict[str, str]:
        assert self.value is not None
        return {self.name: self.value}


class BooleanPredicate(PredicateExpression):
    __slots__ = ("operator", "left_operand", "right_operand")

    def __init__(
        self,
        operator: str,
        left_operand: PredicateExpression,
        right_operand: PredicateExpression,
    ):
        self.operator = operator
        self.left_operand = left_operand
        self.right_operand = right_operand

    def __str__(self):
        sep = " " if self.operator in ("and", "or") else ""
        return f"({self.left_operand}{sep}{self.operator}{sep}{self.right_operand})"

    @staticmethod
    def from_partitions(partitions: List[str]) -> PredicateExpression:
        """
        Instantiates a boolean predicate from a list of string expressions. If the
        expression list passed contains only 1 element, an attribute or index predicate
        object gets instantiated.

        >>> bp = BooleanPredicate.from_partitions(['@foo="1"', 'and', '@bar="2"'])
        >>> bp.operator
        'and'

        >>> isinstance(bp.left_operand, AttributePredicate)
        True

        """
        if len(partitions) == 1:
            return PredicateExpression.parse(partitions[0])

        for operator in BOOLEAN_OPERATORS:
            for i, token in tuple(enumerate(partitions))[-2::-2]:
                if token == operator:
                    return BooleanPredicate(
                        operator,
                        BooleanPredicate.from_partitions(partitions[:i]),
                        BooleanPredicate.from_partitions(partitions[i + 1 :]),
                    )

        raise RuntimeError

    @property
    def _can_describe_attributes(self) -> bool:
        return (
            self.operator == "and"
            and self.left_operand._can_describe_attributes
            and self.right_operand._can_describe_attributes
        )

    def _derive_attributes(self) -> Dict[str, str]:
        if self.operator != "and":
            raise RuntimeError

        return {
            **self.left_operand._derive_attributes(),
            **self.right_operand._derive_attributes(),
        }


class FunctionExpression(PredicateExpression):
    __slots__ = ("arguments", "name")

    def __init__(self, expression: str):
        self.name, rest = expression.split("(", maxsplit=1)
        assert rest.endswith(")")
        # FIXME this fails where an argument itself is a function call,
        #       _split would have to consider brackets as it does with quotes
        self.arguments = tuple(
            PredicateExpression.parse(e) for e in _split(rest[:-1], ",")
        )

    def __str__(self):
        return f"{self.name}({','.join(str(a) for a in self.arguments)})"


class IndexPredicate(PredicateExpression):
    __slots__ = ("index",)

    def __init__(self, index: str):
        self.index = int(index)

    def __str__(self):
        return str(self.index)


class Literal(PredicateExpression):
    __slots__ = ("value",)

    def __init__(self, expression: str):
        assert expression.endswith(expression[0])
        self.value = expression[1:-1]

    def __str__(self):
        value = self.value
        quote = "'" if '"' in value else '"'
        return f"{quote}{value}{quote}"


class UnsupportedPredicate(PredicateExpression):
    __slots__ = ("expression",)

    def __init__(self, expression: str):
        self.expression = expression

    def __str__(self):
        return self.expression



# Copyright (C) 2018-'21  Frank Sachsenheim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import Any, Dict, Union

from _delb.typing import Loader


class FailedDocumentLoading(Exception):
    def __init__(self, source: Any, excuses: Dict[Loader, Union[str, Exception]]):
        self.source = source
        self.excuses = excuses

    def __str__(self):
        return f"Couldn't load {self.source!r} with these loaders: {self.excuses}"


class InvalidCodePath(RuntimeError):
    """
    Raised when a code path that is not expected to be executed is reached.
    """

    def __init__(self):  # pragma: no cover
        super().__init__(
            "An unintended path was taken through the code. Please report this bug."
        )


class InvalidOperation(Exception):
    """
    Raised when an invalid operation is attempted by the client code.
    """

    pass


__all__ = (
    FailedDocumentLoading.__name__,
    InvalidCodePath.__name__,
    InvalidOperation.__name__,
)



# Copyright (C) 2018-'21  Frank Sachsenheim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from inspect import isclass
from types import SimpleNamespace
from typing import Any, Callable, Dict, Iterable, Type
from warnings import warn

import pkg_resources

from _delb.typing import Loader, LoaderConstraint


class DocumentMixinHooks:
    """
    This class acts as termination for methods that can be implemented by mixin classes.
    Any implementation of a method must call a base class' one with :func:`super`.
    """

    @classmethod
    def _init_config(cls, config: SimpleNamespace, kwargs: Dict[str, Any]):
        """
        The ``kwargs`` argument contains the additional keyword arguments that a
        :class:`Document` instance is called with. Extension classes that expect
        configuration data *must* process their specific arguments by clearing them
        from the ``kwargs`` dictionary, e.g. with :meth:`dict.pop`, and preferably
        storing the final configuration data in a :class:`types.SimpleNamespace` and
        adding it to the :class:`types.SimpleNamespace` passed as ``config`` with the
        extension's name. The initially mentioned keyword arguments *should* be prefixed
        with that name as well. This method is called before the loaders try to read and
        parse the given source for a document.
        """
        if kwargs:
            raise RuntimeError(
                "Not all configuration arguments have been processed. You either "
                "passed invalid arguments or an extension doesn't handle them "
                f"properly: {config}"
            )


class DocumentExtensionHooks(DocumentMixinHooks):
    def __init__(self, *args, **kwargs):
        warn(
            "DocumentExtensionHooks is deprecated, use DocumentMixinHooks instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)


class PluginManager:
    def __init__(self):
        self.plugins = SimpleNamespace(
            document_extensions=[],
            loaders=[],
        )

    @staticmethod
    def load_plugins():
        """
        Loads all modules that are registered as entrypoint in the ``delb`` group.
        """
        for entrypoint in pkg_resources.iter_entry_points("delb"):
            entrypoint.load()

    def register_document_mixin(self, extension: Type) -> Type:
        """
        This decorator registers document extension classes which are supposed to add
        additional attributes to a document, e.g. derived data or methods to interact
        with storage systems. All attributes of an extension should share a common
        prefix that terminates with an underscore, e.g. `storage_load`, `storage_save`,
        etc.

        There are hook methods that an extension can implement, they are declared in
        :class:`_delb.plugins.DocumentMixinHooks`.
        """

        assert isclass(extension)
        self.plugins.document_extensions.append(extension)
        return extension

    def register_document_extension(self, extension: Type) -> Type:
        warn(
            "register_document_extension is deprecated, use register_document_mixin "
            "instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.register_document_mixin(extension)

    def register_loader(
        self, before: LoaderConstraint = None, after: LoaderConstraint = None
    ) -> Callable:
        """
        Registers a document loader.

        An example module that is specified as ``delb`` plugin for an IPFS loader might
        look like this:

        .. testcode::

            from os import getenv
            from types import SimpleNamespace
            from typing import Any

            from _delb.plugins import plugin_manager
            from _delb.plugins.https_loader import https_loader
            from _delb.typing import LoaderResult


            IPFS_GATEWAY = getenv("IPFS_GATEWAY_PREFIX", "https://ipfs.io/ipfs/")


            @plugin_manager.register_loader()
            def ipfs_loader(source: Any, config: SimpleNamespace) -> LoaderResult:
                if isinstance(source, str) and source.startswith("ipfs://"):

                    config.source_url = source
                    config.ipfs_gateway_source_url = IPFS_GATEWAY + source[7:]

                    return https_loader(config.ipfs_gateway_source_url, config)

                # return an indication why this loader didn't attempt to load in order
                # to support debugging
                return "The input value is not an URL with the ipfs scheme."


        The ``source`` argument is what a :class:`Document` instance is initialized with
        as input data.

        Note that the ``config`` argument that is passed to a loader function contains
        configuration data, it's the :attr:`delb.Document.config` property after
        :meth:`_init_config <_delb.plugins.DocumentMixinHooks._init_config>` has
        been processed.

        Loaders that retrieve a document from an URL should add the origin as string to
        the ``config`` object as ``source_url``.

        You might want to specify a loader to be considered before or after another
        one. Let's assume a loader shall figure out what to load from a remote XML
        resource that contains a reference to the actual document.
        That one would have to be considered before the one that loads XML documents
        from a URL with the `https` scheme:

        .. testcode::

            from _delb.plugins import plugin_manager
            from _delb.plugins.https_loader import https_loader


            @plugin_manager.register_loader(before=https_loader)
            def mets_loader(source, config) -> LoaderResult:
                # loading logic here
                pass
        """

        if before is not None and after is not None:
            raise NotImplementedError(
                "Loaders may only define one constraint atm. Please open an issue with "
                "a use-case description if you need to define both."
            )

        registered_loaders = self.plugins.loaders

        if before is not None:
            if not isinstance(before, Iterable):
                before = (before,)
            index = min(registered_loaders.index(x) for x in before)

        elif after is not None:
            if not isinstance(after, Iterable):
                after = (after,)
            index = max(registered_loaders.index(x) for x in after) + 1

        else:
            index = len(registered_loaders)

        def registrar(loader: Loader) -> Loader:
            assert callable(loader)
            registered_loaders.insert(index, loader)
            return loader

        return registrar


plugin_manager = PluginManager()



# Copyright (C) 2018-'21  Frank Sachsenheim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
The ``core_loaders`` module provides a set loaders to retrieve documents from various
data sources.
"""

from __future__ import annotations

from copy import deepcopy
from io import IOBase, UnsupportedOperation
from pathlib import Path
from types import SimpleNamespace
from typing import cast, Any, IO

from lxml import etree

from _delb import utils
from _delb.nodes import TagNode
from _delb.plugins import plugin_manager
from _delb.typing import LoaderResult


def tag_node_loader(data: Any, config: SimpleNamespace) -> LoaderResult:
    """
    This loader loads, or rather clones, a :class:`delb.TagNode` instance and its
    descendant nodes.
    """
    if isinstance(data, TagNode):
        tree = etree.ElementTree()
        root = data.clone(deep=True)
        tree._setroot(root._etree_obj)
        utils._copy_root_siblings(data._etree_obj, root._etree_obj)
        return tree
    return "The input value is not a TagNode instance."


@plugin_manager.register_loader()
def etree_loader(data: Any, config: SimpleNamespace) -> LoaderResult:
    """
    This loader processes :class:`lxml.etree._Element` and
    :class:`lxml.etree._ElementTree` instances.
    """
    if isinstance(data, etree._ElementTree):
        return deepcopy(data)
    if isinstance(data, etree._Element):
        return etree.ElementTree(element=deepcopy(data), parser=config.parser)
    return "The input value is neither an etree.Element or …Tree instance."


@plugin_manager.register_loader(after=etree_loader)
def path_loader(data: Any, config: SimpleNamespace) -> LoaderResult:
    """
    This loader loads from a file that is pointed at with a :class:`pathlib.Path`
    instance. That instance will be bound to ``source_path`` on the document's
    :attr:`Document.config` attribute.
    """
    if isinstance(data, Path):
        config.source_url = f"file://{data}"
        with data.open("r") as file:
            return buffer_loader(file, config)
    return "The input value is not a pathlib.Path instance."


@plugin_manager.register_loader(after=path_loader)
def buffer_loader(data: Any, config: SimpleNamespace) -> LoaderResult:
    """
    This loader loads a document from a :term:`file-like object`.
    """
    if isinstance(data, IOBase):
        try:
            data.seek(0)
        except UnsupportedOperation:
            pass
        return etree.parse(cast(IO, data), parser=config.parser)
    return "The input value is no buffer object."


@plugin_manager.register_loader(after=buffer_loader)
def ftp_http_loader(data: Any, config: SimpleNamespace) -> LoaderResult:
    """
    Loads a document from a URL with either the ``ftp`` or ``http`` schema. The URL
    will be bound to ``source_url`` on the document's :attr:`Document.config` attribute.
    """
    if isinstance(data, str) and data.lower().startswith(("http://", "ftp://")):
        config.source_url = data
        return etree.parse(data, parser=config.parser)
    return "The input value is not an URL with the ftp or http scheme."


@plugin_manager.register_loader(after=ftp_http_loader)
def text_loader(data: Any, config: SimpleNamespace) -> LoaderResult:
    """
    Parses a string containing a full document.
    """
    if isinstance(data, str):
        data = data.encode()
    if isinstance(data, bytes):
        root = etree.fromstring(data, config.parser)
        return etree.ElementTree(element=root)
    return "The input value is not a byte sequence."


__all__ = (
    buffer_loader.__name__,
    etree_loader.__name__,
    ftp_http_loader.__name__,
    path_loader.__name__,
    tag_node_loader.__name__,
    text_loader.__name__,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['_delb', '_delb.plugins', 'delb']

package_data = \
{'': ['*']}

install_requires = \
['cssselect', 'lxml', 'setuptools']

extras_require = \
{'https-loader': ['requests>=2.21,<3.0']}

entry_points = \
{'delb': ['https-loader = _delb.plugins.https_loader']}

setup_kwargs = {
    'name': 'delb',
    'version': '0.3.0',
    'description': 'A library that provides an ergonomic model for XML encoded text documents (e.g. with TEI-XML).',
    'long_description': "delb\n====\n\n|latest-version| |rtd| |python-support| |license| |black|\n\n``delb`` is a library that provides an ergonomic model for XML encoded text\ndocuments (e.g. TEI-XML_) for the Python programming language.\nIt fills a gap for the humanities-related field of software development towards\nthe excellent (scientific) communities in the Python ecosystem.\n\nFor a more elaborated discussion see the *Design* chapter of the documentation.\n\n.. _TEI-XML: https://tei-c.org\n\n\nFeatures\n--------\n\n- Loads documents from various source types. This is customizable and\n  extensible.\n- XML DOM types are represented by distinct classes.\n- A completely type-annotated API.\n- Consistent design regarding names and callables' signatures.\n- Shadows comments and processing instructions by default.\n- Querying with XPath and CSS expressions.\n- Applying XSL Transformations.\n\n\nDevelopment status\n------------------\n\nYou're invited to submit tests that reflect desired use cases or are merely of\ntheoretical nature. Of course, any kind of proposals for or implementations of\nimprovements are welcome as well.\n\n\nRelated Projects & Testimonials\n-------------------------------\n\nsnakesist_ is an eXist-db client that uses ``delb`` to expose database\nresources.\n\nKurt Raschke `noted in 2010`_::\n\n  In a DOM-based implementation, it would be relatively easy […]\n  But lxml doesn't use text nodes; instead it uses and properties to hold text\n  content.\n\n\n.. _snakesist: https://pypi.org/project/snakesist/\n.. _noted in 2010: https://web.archive.org/web/20190316214219/https://kurtraschke.com/2010/09/lxml-inserting-elements-in-text/\n\n\n\n.. |black| image:: https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square\n   :alt: Black code style\n   :target: https://black.readthedocs.io/\n.. |latest-version| image:: https://img.shields.io/pypi/v/delb.svg?style=flat-square\n   :alt: Latest version on PyPI\n   :target: https://pypi.org/project/delb\n.. |license| image:: https://img.shields.io/pypi/l/delb.svg?style=flat-square\n   :alt: License\n   :target: https://github.com/funkyfuture/delb/blob/main/LICENSE.txt\n.. |python-support| image:: https://img.shields.io/pypi/pyversions/delb.svg?style=flat-square\n   :alt: Python versions\n.. |rtd| image:: https://img.shields.io/badge/RTD-Docs-informational.svg?style=flat-square\n   :alt: Documentation\n   :target: https://delb.readthedocs.io/\n",
    'author': 'Frank Sachsenheim',
    'author_email': 'funkyfuture@riseup.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/funkyfuture/delb',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
